#!/usr/sepp/bin/perl -w

# mailgraph -- a postfix statistics rrdtool frontend
# copyright (c) 2000, David Schweikert <dws@ee.ethz.ch>
# released under the GNU General Public License

use lib "/usr/pack/rrdtool-1.0.28-mo/lib/perl";
use lib "/home/dws/lib/site_perl/5.6.0";
use lib "/home/dws/lib/site_perl/5.6.0/sun4-solaris/";

use RRDs;
use Time::Local;

use strict;
use File::Tail;
use Getopt::Long;

my $VERSION = 0.11;

# config
my $rrdstep = 60;
my $xpoints = 540;
my $points_per_sample = 3;
my $ypoints = 150;
my $ypoints_err = 80;
my $graph_period = 1800; # 30 minutes
my $graph_mintime = 20; # 20 seconds

# global variables
my $logfile;
my $rrd; # rrd file
my $year;
my $this_minute;
my %sum = ( sent => 0, received => 0, bounced => 0, rejected => 0 );
my $rrd_inited=0;
my $syslog_re = qr'^(...)[ ]{1,2}(\d+) (\d+):(\d+):(\d+) (\S+) (\S+?): (.*)'o;
my $last_graphed=0;

# prototypes
sub process_line($);
sub syslogtime2time($$$$$);
sub event_sent($);
sub event_received($);
sub event_bounced($);
sub event_rejected($);
sub init_rrd($);
sub update($);
sub graph();

sub usage
{
	print "usage: mailgraph [*options*]\n\n";
	print "  -c, --cat          causes the logfile to be only read and not monitored\n";
	print "  -l, --logfile f    monitor logfile f instead of /var/log/syslog\n";
	print "  -g, --graph        generate the graphs and exit\n";
	print "  -h, --help         display this help and exit\n";
	print "  -v, --version      output version information and exit\n";

	exit;
}

sub main
{
	my %opt = ();
	GetOptions(\%opt, 'help|h', 'cat|c', 'graph|g', 'logfile|l=s', 'version|v') or usage;
	usage if $opt{help};

	if($opt{version}) {
		print "mailgraph $VERSION by dws\@ee.ethz.ch\n";
		exit;
	}

	my $logfile = defined $opt{logfile} ? $opt{logfile} : '/var/log/syslog';

	$rrd = "mailgraph.rrd";

	if($opt{graph}) {
		graph();
		exit;
	}

	if($opt{cat}) {
		open(FILE, "<$logfile") or die "can't open $logfile\n";
		while(<FILE>) {
			process_line($_);
		}
		graph();
	}
	else {
		my $file = File::Tail->new(name=>$logfile, tail=>-1);
		my $line;
		my $graph_minute=0;
		my $graph_time=0;
		while (defined($line=$file->read)) {
			process_line($line);
			$this_minute or next;
			if($this_minute-$graph_minute >= $graph_period) {
				my $now = time;
				next if $now-$graph_time<$graph_mintime;
				graph();
				$graph_minute=$this_minute;
				$graph_time = $now;
			}
		}
	}
}

sub init_rrd($)
{
	my $m = shift;
	if(not -f $rrd) {
		my $rows = $xpoints/$points_per_sample;
		my $realrows = int($rows*1.5); # ensure that the full range is covered
		my $day_steps = int(3600*24 / ($rrdstep*$rows));
		my $week_steps = $day_steps*7;
		my $month_steps = $day_steps*31;
		my $year_steps = $day_steps*365;
		RRDs::create($rrd, '--start', $m, '--step', $rrdstep,
				'DS:sent:ABSOLUTE:'.($rrdstep*2).':0:U',
				'DS:recv:ABSOLUTE:'.($rrdstep*2).':0:U',
				'DS:bounced:ABSOLUTE:'.($rrdstep*2).':0:U',
				'DS:rejected:ABSOLUTE:'.($rrdstep*2).':0:U',
				"RRA:AVERAGE:0.5:$day_steps:$realrows",   # day
				"RRA:AVERAGE:0.5:$week_steps:$realrows",  # week
				"RRA:AVERAGE:0.5:$month_steps:$realrows", # month
				"RRA:AVERAGE:0.5:$year_steps:$realrows",  # year
				"RRA:MAX:0.5:$day_steps:$realrows",   # day
				"RRA:MAX:0.5:$week_steps:$realrows",  # week
				"RRA:MAX:0.5:$month_steps:$realrows", # month
				"RRA:MAX:0.5:$year_steps:$realrows",  # year
				);
		$this_minute = $m;
	}
	else {
		$this_minute = RRDs::last($rrd) + $rrdstep;
	}
	$rrd_inited=1;
}

sub syslogtime2time($$$$$)
{

	my ($mstr,$d,$H,$M,$S) = @_;
	my %months = (
		'Jan' => 0, 'Feb' =>  1, 'Mar' =>  2,
		'Apr' => 3, 'May' =>  4, 'Jun' =>  5,
		'Jul' => 6, 'Aug' =>  7, 'Sep' =>  8,
		'Oct' => 9, 'Nov' => 10, 'Dec' => 11 );
	my $m = $months{$mstr};
	if(not defined $m) {
		warn "Can't parse month '$1'\n";
		return;
	};

	$year = (localtime(time))[5] unless defined $year;
	my $time = timelocal($S,$M,$H,$d,$m,$year);
	if(defined $this_minute and $time < $this_minute) {
		# Dec 31 -> Jan  1 
		$year = (localtime(time))[5];
		$time = timelocal($S,$M,$H,$d,$m,$year);
	}
	return $time;
}

sub process_line($)
{
	my $line = shift;
	# parse syslog syntax
	$line =~ /$syslog_re/o or return;
	my ($m,$d,$H,$M,$S) = ($1,$2,$3,$4,$5);
	my $host = $6;
	my $prog = $7;
	my $str = $8;
	$prog =~ /^postfix\/(.*)/ or return; $prog = $1;
	my $time = syslogtime2time($m,$d,$H,$M,$S);
	$prog =~ s/\[\d+\]//; # strip process number
	if($prog eq 'smtp') {
		if($str =~ /\bstatus=sent\b/) {
			event_sent($time);
		}
		if($str =~ /\bstatus=bounced\b/) {
			event_bounced($time);
		}
	}
	elsif($prog eq 'local') {
		if($str =~ /\bstatus=bounced\b/) {
			event_bounced($time);
		}
	}
	elsif($prog eq 'smtpd') {
		if($str =~ /^[0-9A-F]+: client=/) {
			event_received($time);
		}
		if($str =~ /^reject: /) {
			event_rejected($time);
		}
	}
}

sub event_sent($)
{
	my $t = shift;
	update($t) and $sum{sent}++;
}

sub event_received($)
{
	my $t = shift;
	update($t) and $sum{received}++;
}

sub event_bounced($)
{
	my $t = shift;
	update($t) and $sum{bounced}++;
}

sub event_rejected($)
{
	my $t = shift;
	update($t) and $sum{rejected}++;
}

# returns 1 if $sum should be updated
sub update($)
{
	my $t = shift;
	my $m = $t - $t%$rrdstep;
	init_rrd($m) unless $rrd_inited;
	return 1 if $m == $this_minute;
	return 0 if $m < $this_minute;

	#print "$m\n";
	print "update $this_minute:$sum{sent}:$sum{received}:$sum{bounced}:$sum{rejected}\n";
	RRDs::update $rrd, "$this_minute:$sum{sent}:$sum{received}:$sum{bounced}:$sum{rejected}";
	if($m > $this_minute+$rrdstep) {
		for(my $sm=$this_minute+$rrdstep;$sm<$m;$sm+=$rrdstep) {
			print "update $sm:0:0:0:0 (SKIP)\n";
			RRDs::update $rrd, "$sm:0:0:0:0";
		}
	}
	$this_minute = $m;
	$sum{sent}=0;
	$sum{received}=0;
	$sum{bounced}=0;
	$sum{rejected}=0;
	return 1;
}

sub make_graph($$$)
{
	my $range = shift;
	my $file = shift;
	my $title = shift;

	RRDs::graph("$file.png",
		'--imgformat', 'PNG',
		'--width', $xpoints,
		'--height', $ypoints,
		'--start', "-$range",
		'--vertical-label', 'msgs/hour',
		'--title', $title,
		"DEF:sent=$rrd:sent:AVERAGE",
		"DEF:recv=$rrd:recv:AVERAGE",
		"DEF:msent=$rrd:sent:MAX",
		"DEF:mrecv=$rrd:recv:MAX",
		"CDEF:rsent=sent,60,*",
		"CDEF:rrecv=recv,60,*",
		"CDEF:vsent=sent,$range,*",
		"CDEF:vrecv=recv,$range,*",
		"CDEF:rmsent=msent,60,*",
		"CDEF:rmrecv=mrecv,60,*",
		'AREA:rsent#000099:Sent',
		'GPRINT:vsent:AVERAGE:total\: %.0lf msgs',
		'GPRINT:rmsent:MAX:max\: %.0lf msgs/min',
		'LINE2:rrecv#009900:Received',
		'GPRINT:vrecv:AVERAGE:total\: %.0lf msgs',
		'GPRINT:rmrecv:MAX:max\: %.0lf msgs/min\l',
		'HRULE:0#000000',
		'COMMENT:\s',
		'COMMENT:Graph Created on '.localtime(time).'\r',
	);
		#"COMMENT:".localtime($this_minute-$rrdstep),
	not RRDs::error or die "ERROR while graphing $file: ".RRDs::error."\n";

	RRDs::graph("${file}_err.png",
		'--imgformat', 'PNG',
		'--width', $xpoints,
		'--height', $ypoints_err,
		'--start', "-$range",
		'--vertical-label', 'msgs/hour',
		'--title', "$title of Errors",
		"DEF:bounced=$rrd:bounced:AVERAGE",
		"DEF:rejected=$rrd:rejected:AVERAGE",
		"DEF:mbounced=$rrd:bounced:MAX",
		"DEF:mrejected=$rrd:rejected:MAX",
		"CDEF:rbounced=bounced,3600,*",
		"CDEF:rrejected=rejected,3600,*",
		"CDEF:vbounced=bounced,$range,*",
		"CDEF:vrejected=rejected,$range,*",
		"CDEF:rmbounced=mbounced,60,*",
		"CDEF:rmrejected=mrejected,60,*",
		'AREA:rbounced#000000:Bounced',
		'GPRINT:vbounced:AVERAGE:total\: %.0lf msgs',
		'GPRINT:rmbounced:MAX:max\: %.0lf msgs/min',
		'LINE2:rrejected#BB0000:Rejected',
		'GPRINT:vrejected:AVERAGE:total\: %.0lf msgs',
		'GPRINT:rmrejected:MAX:max\: %.0lf msgs/min\l',
		'HRULE:0#000000',
		'COMMENT:\s',
		'COMMENT:Graph Created on '.localtime(time).'\r',
	);
	not RRDs::error or die "ERROR while graphing $file: ".RRDs::error."\n";
}

sub graph()
{
	print "generating graphs...\n";

	my $day = 3600*24;
	my $week = $day*7;
	my $month = $day*31;
	my $year = $day*365;

	make_graph($day, "mailgraph_day", "Day Graph");
	make_graph($week, "mailgraph_week", "Week Graph");
	make_graph($month, "mailgraph_month", "Month Graph");
	make_graph($year-$day, "mailgraph_year", "Year Graph");
}

main;
